@@ -1,9 +1,10 @@
 # <@LICENSE>
-# Copyright 2004 Apache Software Foundation
-# 
-# Licensed under the Apache License, Version 2.0 (the "License");
-# you may not use this file except in compliance with the License.
-# You may obtain a copy of the License at
+# Licensed to the Apache Software Foundation (ASF) under one or more
+# contributor license agreements.  See the NOTICE file distributed with
+# this work for additional information regarding copyright ownership.
+# The ASF licenses this file to you under the Apache License, Version 2.0
+# (the "License"); you may not use this file except in compliance with
+# the License.  You may obtain a copy of the License at:
 # 
 #     http://www.apache.org/licenses/LICENSE-2.0
 # 
@@ -57,7 +58,7 @@ The type of this setting:
 
            - $CONF_TYPE_STRING: string
            - $CONF_TYPE_NUMERIC: numeric value (float or int)
-           - $CONF_TYPE_BOOL: boolean (0 or 1)
+           - $CONF_TYPE_BOOL: boolean (0/no or 1/yes)
            - $CONF_TYPE_TEMPLATE: template, like "report"
            - $CONF_TYPE_ADDRLIST: address list, like "whitelist_from"
            - $CONF_TYPE_HASH_KEY_VALUE: hash key/value pair,
@@ -102,7 +103,8 @@ Set to 1 if this setting requires 'allow_user_rules' when run from spamd.
 =item is_admin
 
 Set to 1 if this setting can only be set in the system-wide config when run
-from spamd.
+from spamd.  (All settings can be used by local programs run directly by the
+user.)
 
 =item is_frequent
 
@@ -111,9 +113,6 @@ up first for speed.
 
 =back
 
-Note that the registered commands array can be extended by plugins, by adding
-the new config settings to the C<$conf-<gt>{registered_commands}> array ref.
-
 =head1 METHODS
 
 =over 4
@@ -121,12 +120,14 @@ the new config settings to the C<$conf-<gt>{registered_commands}> array ref.
 =cut
 
 package Mail::SpamAssassin::Conf::Parser;
+
 use Mail::SpamAssassin::Conf;
 use Mail::SpamAssassin::Constants qw(:sa);
+use Mail::SpamAssassin::Logger;
 
 use strict;
+use warnings;
 use bytes;
-use Carp;
 
 use vars qw{
   @ISA
@@ -145,16 +146,29 @@ sub new {
     'conf'      => $conf
   };
 
+  $self->{command_luts} = { };
+  $self->{command_luts}->{frequent} = { };
+  $self->{command_luts}->{remaining} = { };
+
   bless ($self, $class);
   $self;
 }
 
 ###########################################################################
 
-sub set_defaults_from_command_list {
-  my ($self) = @_;
+sub register_commands {
+  my($self, $arrref) = @_;
   my $conf = $self->{conf};
-  foreach my $cmd (@{$conf->{registered_commands}}) {
+
+  $self->set_defaults_from_command_list($arrref);
+  $self->build_command_luts($arrref);
+  push(@{$conf->{registered_commands}}, @{$arrref});
+}
+
+sub set_defaults_from_command_list {
+  my ($self, $arrref) = @_;
+  my $conf = $self->{conf};
+  foreach my $cmd (@{$arrref}) {
     # note! exists, not defined -- we want to be able to set
     # "undef" default values.
     if (exists($cmd->{default})) {
@@ -164,19 +178,12 @@ sub set_defaults_from_command_list {
 }
 
 sub build_command_luts {
-  my ($self) = @_;
+  my ($self, $arrref) = @_;
 
-  return if $self->{already_built_config_lookup};
-  $self->{already_built_config_lookup} = 1;
-
-  $self->{command_luts} = { };
-  $self->{command_luts}->{frequent} = { };
-  $self->{command_luts}->{remaining} = { };
   my $conf = $self->{conf};
 
   my $set;
-  foreach my $cmd (@{$conf->{registered_commands}})
-  {
+  foreach my $cmd (@{$arrref}) {
     # first off, decide what set this is in.
     if ($cmd->{is_frequent}) { $set = 'frequent'; }
     else { $set = 'remaining'; }
@@ -184,8 +191,12 @@ sub build_command_luts {
     # next, its priority (used to ensure frequently-used params
     # are parsed first)
     my $cmdname = $cmd->{command} || $cmd->{setting};
-    foreach my $name ($cmdname, @{$cmd->{aliases}}) {
-      $self->{command_luts}->{$set}->{$name} = $cmd;
+    $self->{command_luts}->{$set}->{$cmdname} = $cmd;
+
+    if ($cmd->{aliases} && scalar @{$cmd->{aliases}} > 0) {
+      foreach my $name (@{$cmd->{aliases}}) {
+        $self->{command_luts}->{$set}->{$name} = $cmd;
+      }
     }
   }
 }
@@ -216,10 +227,11 @@ sub parse {
     $lang =~ s/[@.+,].*$//;    # Strip codeset, modifier/audience, etc.
   }                            # (eg. .utf8 or @euro)
 
-  # build and get fast-access handles on the command lookup tables
-  $self->build_command_luts();
+  # get fast-access handles on the command lookup tables
   my $lut_frequent = $self->{command_luts}->{frequent};
   my $lut_remaining = $self->{command_luts}->{remaining};
+  my %migrated_keys = map { $_ => 1 }
+            @Mail::SpamAssassin::Conf::MIGRATED_SETTINGS;
 
   $self->{currentfile} = '(no file)';
   my $skip_parsing = 0;
@@ -227,9 +239,13 @@ sub parse {
   my @if_stack = ();
   my @conf_lines = split (/\n/, $_[1]);
   my $line;
+  $self->{if_stack} = \@if_stack;
 
   while (defined ($line = shift @conf_lines)) {
+    local ($1);         # bug 3838: prevent random taint flagging of $1
+
     $line =~ s/(?<!\\)#.*$//; # remove comments
+    $line =~ s/\\#/#/g; # hash chars are escaped, so unescape them
     $line =~ s/^\s+//;  # remove leading whitespace
     $line =~ s/\s+$//;  # remove tailing whitespace
     next unless($line); # skip empty lines
@@ -262,14 +278,17 @@ sub parse {
           my $cond = pop @if_stack;
 
           if ($cond->{type} eq 'if') {
-            warn "unclosed 'if' in ".
+            my $msg = "config: unclosed 'if' in ".
                   $self->{currentfile}.": if ".$cond->{conditional}."\n";
+            warn $msg;
+            $self->lint_warn($msg, undef);
           }
           else {
-            die "unknown 'if' type: ".$cond->{type}."\n";
+            # die seems a bit excessive here, but this shouldn't be possible
+            # so I suppose it's okay.
+            die "config: unknown 'if' type: ".$cond->{type}."\n";
           }
 
-          $conf->{errors}++;
           @if_stack = ();
         }
         $skip_parsing = 0;
@@ -304,9 +323,25 @@ sub parse {
       next;
     }
 
+    if ($key eq 'else') {
+      # TODO: if/else/else won't get flagged here :(
+      if (!@if_stack) {
+        $parse_error = "config: found else without matching conditional";
+        goto failed_line;
+      }
+
+      $skip_parsing = !$skip_parsing;
+      next;
+    }
+
     # and the endif statement:
     if ($key eq 'endif') {
       my $lastcond = pop @if_stack;
+      if (!defined $lastcond) {
+        $parse_error = "config: found endif without matching conditional";
+        goto failed_line;
+      }
+
       $skip_parsing = $lastcond->{skip_parsing};
       next;
     }
@@ -329,13 +364,14 @@ sub parse {
       #$value =~ s/^(\d+)\.(\d{1,3}).*$/sprintf "%d.%d", $1, $2/e;
 
       if ($ver ne $value) {
-        warn "configuration file \"$self->{currentfile}\" requires version ".
-                "$value of SpamAssassin, but this is code version ".
+        my $msg = "config: configuration file \"$self->{currentfile}\" requires ".
+                "version $value of SpamAssassin, but this is code version ".
                 "$ver. Maybe you need to use ".
                 "the -C switch, or remove the old config files? ".
                 "Skipping this file";
+        warn $msg;
+        $self->lint_warn($msg, undef);
         $skip_parsing = 1;
-        $conf->{errors}++;
       }
       next;
     }
@@ -350,17 +386,19 @@ sub parse {
     if ($cmd) {
       if ($self->{scoresonly}) {              # reading user config from spamd
         if ($cmd->{is_priv} && !$conf->{allow_user_rules}) {
-          dbg ("config: not parsing, 'allow_user_rules' is 0: $line");
+          info("config: not parsing, 'allow_user_rules' is 0: $line");
           goto failed_line;
         }
         if ($cmd->{is_admin}) {
-          dbg ("config: not parsing, administrator setting: $line");
+          info("config: not parsing, administrator setting: $line");
           goto failed_line;
         }
       }
 
       if (!$cmd->{code}) {
-        $self->setup_default_code_cb ($cmd);
+        if (! $self->setup_default_code_cb($cmd)) {
+          goto failed_line;
+        }
       }
 
       my $ret = &{$cmd->{code}} ($conf, $cmd->{setting}, $value, $line);
@@ -385,7 +423,7 @@ sub parse {
     }
 
     # last ditch: try to see if the plugins know what to do with it
-    if ($conf->{main}->call_plugins ("parse_config", {
+    if ($conf->{main}->call_plugins("parse_config", {
                 key => $key,
                 value => $value,
                 line => $line,
@@ -399,20 +437,24 @@ sub parse {
 
 failed_line:
     my $msg = $parse_error;
+    my $is_error = 1;
     if (!$msg) {
-      # the default warning, if a more specific one isn't output
-      $msg = "config: SpamAssassin failed to parse line, ".
-                        "skipping: $line";
+      # use a default warning, if a more specific one wasn't output
+      if ($migrated_keys{$key}) {
+        # this key was moved into a plugin; non-fatal for lint
+        $is_error = 0;
+        $msg = "config: failed to parse, now a plugin, skipping, in \"$self->{currentfile}\": $line";
+      } else {
+        # a real syntax error; this is fatal for --lint
+        $msg = "config: failed to parse line, skipping, in \"$self->{currentfile}\": $line";
+      }
     }
 
-    if ($conf->{lint_rules}) {
-      warn $msg."\n";
-    } else {
-      dbg ($msg);
-    }
-    $conf->{errors}++;
+    $self->lint_warn($msg, undef, $is_error);
   }
 
+  delete $self->{if_stack};
+
   $self->lint_check();
   $self->set_default_scores();
 
@@ -429,7 +471,7 @@ sub handle_conditional {
   my $eval = '';
   my $bad = 0;
   foreach my $token (@tokens) {
-    if ($token =~ /^(\W+|[\-\+\d\.]+)$/) {
+    if ($token =~ /^(\W+|[+-]?\d+(?:\.\d+)?)$/) {
       $eval .= $1." ";          # note: untaints!
     }
     elsif ($token eq 'plugin') {
@@ -443,12 +485,13 @@ sub handle_conditional {
       $eval .= "\"$1\" ";       # note: untaints!
     }
     else {
-      $bad++; warn "unparseable chars in 'if $value': '$token'\n";
+      $bad++;
+      warn "config: unparseable chars in 'if $value': '$token'\n";
     }
   }
 
   if ($bad) {
-    $conf->{errors}++;
+    $self->lint_warn("bad 'if' line, in \"$self->{currentfile}\"", undef);
     return -1;
   }
 
@@ -478,26 +521,19 @@ sub cond_clause_plugin_loaded {
 sub lint_check {
   my ($self) = @_;
   my $conf = $self->{conf};
-  my ($k, $v);
 
-  if ($conf->{lint_rules})
-  {
+  if ($conf->{lint_rules}) {
     # Check for description and score issues in lint fashion
-    while ( ($k,$v) = each %{$conf->{descriptions}} ) {
-      if (length($v) > 50) {
-        warn "warning: description for $k is over 50 chars\n";
-        $conf->{errors}++;
-      }
+    while ( my $k = each %{$conf->{descriptions}} ) {
       if (!exists $conf->{tests}->{$k}) {
-        warn "warning: description exists for non-existent rule $k\n";
-        $conf->{errors}++;
+        $self->lint_warn("config: warning: description exists for non-existent rule $k\n", $k);
       }
     }
 
     while ( my($sk) = each %{$conf->{scores}} ) {
       if (!exists $conf->{tests}->{$sk}) {
-        warn "warning: score set for non-existent rule $sk\n";
-        $conf->{errors}++;
+        # bug 5514: not a lint warning any more
+        dbg("config: warning: score set for non-existent rule $sk");
       }
     }
   }
@@ -511,21 +547,13 @@ sub lint_check {
 sub set_default_scores {
   my ($self) = @_;
   my $conf = $self->{conf};
-  my ($k, $v);
-
-  while ( ($k,$v) = each %{$conf->{tests}} ) {
-    if ($conf->{lint_rules}) {
-      if (length($k) > 22 && $k !~ /^__/ && $k !~ /^T_/) {
-        warn "warning: rule '$k' is over 22 chars\n";
-        $conf->{errors}++;
-      }
-    }
 
+  while ( my $k = each %{$conf->{tests}} ) {
     if ( ! exists $conf->{scores}->{$k} ) {
       # T_ rules (in a testing probationary period) get low, low scores
       my $set_score = ($k =~/^T_/) ? 0.01 : 1.0;
 
-      $set_score = -$set_score if ( $conf->{tflags}->{$k} =~ /\bnice\b/ );
+      $set_score = -$set_score if ( ($conf->{tflags}->{$k}||'') =~ /\bnice\b/ );
       for my $index (0..3) {
         $conf->{scoreset}->[$index]->{$k} = $set_score;
       }
@@ -558,14 +586,19 @@ sub setup_default_code_cb {
     $cmd->{code} = \&set_template_append;
   }
   else {
-    die "unknown conf type $type!";
+    warn "config: unknown conf type $type!";
+    return 0;
   }
+  return 1;
 }
 
 sub set_numeric_value {
   my ($conf, $key, $value, $line) = @_;
 
-  unless (defined $value && $value =~ /^-?\d+(?:\.\d+)?$/) {
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
+  unless ($value =~ /^-?\d+(?:\.\d+)?$/) {
     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
   }
 
@@ -575,7 +608,20 @@ sub set_numeric_value {
 sub set_bool_value {
   my ($conf, $key, $value, $line) = @_;
 
-  unless (defined $value && ($value == 1 || $value == 0) ) {
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
+
+  # bug 4462: allow yes/1 and no/0 for boolean values
+  $value = lc $value;
+  if ($value eq 'yes') {
+    $value = 1;
+  }
+  elsif ($value eq 'no') {
+    $value = 0;
+  }
+
+  unless ($value =~ /^[01]$/) {
     return $Mail::SpamAssassin::Conf::INVALID_VALUE;
   }
 
@@ -585,7 +631,7 @@ sub set_bool_value {
 sub set_string_value {
   my ($conf, $key, $value, $line) = @_;
 
-  unless (defined $value) {
+  unless (defined $value && $value !~ /^$/) {
     return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
   }
 
@@ -605,11 +651,19 @@ sub set_hash_key_value {
 
 sub set_addrlist_value {
   my ($conf, $key, $value, $line) = @_;
+
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
   $conf->{parser}->add_to_addrlist ($key, split (' ', $value));
 }
 
 sub remove_addrlist_value {
   my ($conf, $key, $value, $line) = @_;
+
+  unless (defined $value && $value !~ /^$/) {
+    return $Mail::SpamAssassin::Conf::MISSING_REQUIRED_VALUE;
+  }
   $conf->{parser}->remove_from_addrlist ($key, split (' ', $value));
 }
 
@@ -626,11 +680,24 @@ sub set_template_clear {
 
 ###########################################################################
 
-# note: error 70 == SA_SOFTWARE
 sub finish_parsing {
-  my ($self) = @_;
+  my ($self, $isuserconf) = @_;
   my $conf = $self->{conf};
 
+  # note: this function is called once for system-wide configuration
+  # with $isuserconf set to 0, then again for user conf with $isuserconf set to 1.
+
+  $self->trace_meta_dependencies();
+  $self->fix_priorities();
+
+  # don't do this if allow_user_rules is active, since it deletes entries
+  # from {tests}
+  if (!$conf->{allow_user_rules}) {
+    $self->find_dup_rules();          # must be after fix_priorities()
+  }
+
+  dbg("conf: finish parsing");
+
   while (my ($name, $text) = each %{$conf->{tests}}) {
     my $type = $conf->{test_types}->{$name};
     my $priority = $conf->{priority}->{$name} || 0;
@@ -638,39 +705,39 @@ sub finish_parsing {
 
     # eval type handling
     if (($type & 1) == 1) {
-      my @args;
       if (my ($function, $args) = ($text =~ m/(.*?)\s*\((.*?)\)\s*$/)) {
-        if ($args) {
-          @args = ($args =~ m/['"](.*?)['"]\s*(?:,\s*|$)/g);
+        my ($packed, $argsref) =
+                $self->pack_eval_method($function, $args, $name, $text);
+
+        if (!$packed) {
+          # we've already warned about this
         }
-        unshift(@args, $function);
-        if ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS) {
-          $conf->{body_evals}->{$priority}->{$name} = \@args;
+        elsif ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_EVALS) {
+          $conf->{body_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_EVALS) {
-          $conf->{head_evals}->{$priority}->{$name} = \@args;
+          $conf->{head_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS) {
           # We don't do priorities for $Mail::SpamAssassin::Conf::TYPE_RBL_EVALS
-          $conf->{rbl_evals}->{$name} = \@args;
+          # we also use the arrayref instead of the packed string
+          $conf->{rbl_evals}->{$name} = [ $function, @$argsref ];
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_EVALS) {
-          $conf->{rawbody_evals}->{$priority}->{$name} = \@args;
+          $conf->{rawbody_evals}->{$priority}->{$name} = $packed;
         }
         elsif ($type == $Mail::SpamAssassin::Conf::TYPE_FULL_EVALS) {
-          $conf->{full_evals}->{$priority}->{$name} = \@args;
+          $conf->{full_evals}->{$priority}->{$name} = $packed;
         }
         #elsif ($type == $Mail::SpamAssassin::Conf::TYPE_URI_EVALS) {
-        #  $conf->{uri_evals}->{$priority}->{$name} = \@args;
+        #  $conf->{uri_evals}->{$priority}->{$name} = $packed;
         #}
         else {
-          $conf->{errors}++;
-          sa_die(70, "unknown type $type for $name: $text");
+          $self->lint_warn("unknown type $type for $name: $text", $name);
         }
       }
       else {
-        $conf->{errors}++;
-        sa_die(70, "syntax error for eval function $name: $text");
+        $self->lint_warn("syntax error for eval function $name: $text", $name);
       }
     }
     # non-eval tests
@@ -682,15 +749,6 @@ sub finish_parsing {
         $conf->{head_tests}->{$priority}->{$name} = $text;
       }
       elsif ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS) {
-        # Meta Tests must have a priority of at least META_TEST_MIN_PRIORITY,
-        # if it's lower then reset the value
-        if ($priority < META_TEST_MIN_PRIORITY) {
-          # we need to lower the count of the old priority and raise the
-          # count of the new priority
-          $conf->{priorities}->{$priority}--;
-          $priority = META_TEST_MIN_PRIORITY;
-          $conf->{priorities}->{$priority}++;
-        }
         $conf->{meta_tests}->{$priority}->{$name} = $text;
       }
       elsif ($type == $Mail::SpamAssassin::Conf::TYPE_URI_TESTS) {
@@ -703,14 +761,233 @@ sub finish_parsing {
         $conf->{full_tests}->{$priority}->{$name} = $text;
       }
       else {
-        $conf->{errors}++;
-        sa_die(70, "unknown type $type for $name: $text");
+        $self->lint_warn("unknown type $type for $name: $text", $name);
       }
     }
   }
 
-  delete $conf->{tests};                # free it up
-  delete $conf->{priority};             # free it up
+  $self->lint_trusted_networks();
+
+  if (!$isuserconf) {
+    # named this way just in case we ever want a "finish_parsing_start"
+    $conf->{main}->call_plugins("finish_parsing_end", { conf => $conf });
+  } else {
+    $conf->{main}->call_plugins("user_conf_parsing_end", { conf => $conf });
+  }
+
+  if (!$conf->{allow_user_rules}) {
+    # free up stuff we no longer need
+    delete $conf->{tests};
+    delete $conf->{priority};
+    delete $conf->{test_types};
+  }
+}
+
+sub trace_meta_dependencies {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+  $conf->{meta_dependencies} = { };
+
+  foreach my $name (keys %{$conf->{tests}}) {
+    next unless ($conf->{test_types}->{$name}
+                    == $Mail::SpamAssassin::Conf::TYPE_META_TESTS);
+
+    my $deps = [ ];
+    my $alreadydone = { };
+    $self->_meta_deps_recurse($conf, $name, $name, $deps, $alreadydone);
+    $conf->{meta_dependencies}->{$name} = join (' ', @{$deps});
+  }
+}
+
+sub _meta_deps_recurse {
+  my ($self, $conf, $toprule, $name, $deps, $alreadydone) = @_;
+
+  # Only do each rule once per top-level meta; avoid infinite recursion
+  return if $alreadydone->{$name};
+  $alreadydone->{$name} = 1;
+
+  # Obviously, don't trace empty or nonexistent rules
+  my $rule = $conf->{tests}->{$name};
+  return unless $rule;
+
+  # Lex the rule into tokens using a rather simple RE method ...
+  my $lexer = ARITH_EXPRESSION_LEXER;
+  my @tokens = ($rule =~ m/$lexer/g);
+
+  # Go through each token in the meta rule
+  foreach my $token (@tokens) {
+    # has to be an alpha+numeric token
+    next if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/);
+    # and has to be a rule name
+    next unless exists $conf->{tests}->{$token};
+
+    # add and recurse
+    push @{$deps}, $token;
+    $self->_meta_deps_recurse($conf, $toprule, $token, $deps, $alreadydone);
+  }
+}
+
+sub fix_priorities {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  die unless $conf->{meta_dependencies};    # order requirement
+  my $pri = $conf->{priority};
+
+  # sort into priority order, lowest first -- this way we ensure that if we
+  # rearrange the pri of a rule early on, we cannot accidentally increase its
+  # priority later.
+  foreach my $rule (sort {
+            $pri->{$a} <=> $pri->{$b}
+          } keys %{$pri})
+  {
+    # we only need to worry about meta rules -- they are the
+    # only type of rules which depend on other rules
+    my $deps = $conf->{meta_dependencies}->{$rule};
+    next unless (defined $deps);
+
+    my $basepri = $pri->{$rule};
+    foreach my $dep (split ' ', $deps) {
+      my $deppri = $pri->{$dep};
+      if ($deppri > $basepri) {
+        dbg("rules: $rule (pri $basepri) requires $dep (pri $deppri): fixed");
+        $pri->{$dep} = $basepri;
+      }
+    }
+  }
+}
+
+sub find_dup_rules {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  my %names_for_text = ();
+  my %dups = ();
+  while (my ($name, $text) = each %{$conf->{tests}}) {
+    my $type = $conf->{test_types}->{$name};
+    my $tf = ($conf->{tflags}->{$name}||''); $tf =~ s/\s+/ /gs;
+    # ensure similar, but differently-typed, rules are not marked as dups;
+    # take tflags into account too due to "tflags multiple"
+    $text = "$type\t$text\t$tf";
+
+    if (defined $names_for_text{$text}) {
+      $names_for_text{$text} .= " ".$name;
+      $dups{$text} = undef;     # found (at least) one
+    } else {
+      $names_for_text{$text} = $name;
+    }
+  }
+
+  foreach my $text (keys %dups) {
+    my $first;
+    my $first_pri;
+    my @names = sort {$a cmp $b} split(' ', $names_for_text{$text});
+    foreach my $name (@names) {
+      my $priority = $conf->{priority}->{$name} || 0;
+
+      if (!defined $first || $priority < $first_pri) {
+        $first_pri = $priority;
+        $first = $name;
+      }
+    }
+    # $first is now the earliest-occurring rule. mark others as dups
+
+    my @dups = ();
+    foreach my $name (@names) {
+      next if $name eq $first;
+      push @dups, $name;
+      delete $conf->{tests}->{$name};
+    }
+
+    dbg("rules: $first merged duplicates: ".join(' ', @dups));
+    $conf->{duplicate_rules}->{$first} = \@dups;
+  }
+}
+
+sub pack_eval_method {
+  my ($self, $function, $args, $name, $text) = @_;
+
+  my @args;
+  if ($args) {
+    # bug 4419: Parse quoted strings, unquoted alphanumerics/floats and
+    # both unquoted IPv4 and IPv6 addresses.  s// is used so that we can
+    # determine whether or not we successfully parsed ALL arguments.
+    while ($args =~ s/^\s*(?:['"](.*?)['"]|([\d\.:A-Za-z]+?))\s*(?:,\s*|$)//) {
+      if (defined $1) {
+        push @args, $1;
+      }
+      else {
+        push @args, $2;
+      }
+    }
+  }
+
+  if ($args) {
+    $self->lint_warn("syntax error (unparsable argument: $args) for eval function: $name: $text", $name);
+    return;
+  }
+
+  my $argstr = $function;
+  $argstr =~ s/\s+//gs;
+
+  if (scalar @args > 0) {
+    $argstr .= ',' . join (', ', map {
+            s/\#/[HASH]/gs; "q#".$_."#" 
+          } @args);
+  }
+  return ($argstr, \@args);
+}
+
+###########################################################################
+
+sub lint_trusted_networks {
+  my ($self) = @_;
+  my $conf = $self->{conf};
+
+  # validate trusted_networks and internal_networks, bug 4760.
+  # check that all internal_networks are listed in trusted_networks
+  # too.  do the same for msa_networks, but check msa_networks against
+  # internal_networks if trusted_networks aren't defined
+
+  my ($nt, $matching_against);
+  if ($conf->{trusted_networks_configured}) {
+    $nt = $conf->{trusted_networks};
+    $matching_against = 'trusted_networks';
+  } elsif ($conf->{internal_networks_configured}) {
+    $nt = $conf->{internal_networks};
+    $matching_against = 'internal_networks';
+  } else {
+    return;
+  }
+
+  foreach my $net_type ('internal_networks', 'msa_networks') {
+    next unless $conf->{"${net_type}_configured"};
+    next if $net_type eq $matching_against;
+
+    my $replace_nets;
+    my @valid_net_list = ();
+    my $net_list = $conf->{$net_type};
+
+    foreach my $net (@{$net_list->{nets}}) {
+      # don't check to see if an excluded network is included - that's senseless
+      if (!$net->{exclude} && !$nt->contains_net($net)) {
+        my $msg = "$matching_against doesn't contain $net_type entry '".
+                  ($net->{as_string})."'";
+
+        $self->lint_warn($msg, undef);      # complain
+        $replace_nets = 1;  # and omit it from the new internal set
+      }
+      else {
+        push @valid_net_list, $net;
+      }
+    }
+
+    if ($replace_nets) {
+      # something was invalid. replace the old nets list with a fixed version
+      # (which may be empty)
+      $net_list->{nets} = \@valid_net_list;
+    }
+  }
 }
 
 ###########################################################################
@@ -720,30 +997,78 @@ sub add_test {
   my $conf = $self->{conf};
 
   # Don't allow invalid names ...
-  if ($name !~ /^\w+$/) {
-    warn "error: rule '$name' has invalid characters (not Alphanumeric + Underscore)\n";
-    $conf->{errors}++;
+  if ($name !~ /^\D\w*$/) {
+    $self->lint_warn("config: error: rule '$name' has invalid characters ".
+	   "(not Alphanumeric + Underscore + starting with a non-digit)\n", $name);
     return;
   }
 
+  # Also set a hard limit for ALL rules (rule names longer than 242
+  # characters throw warnings).  Check this separately from the above
+  # pattern to avoid vague error messages.
+  if (length $name > 200) {
+    $self->lint_warn("config: error: rule '$name' is way too long ".
+	   "(recommended maximum length is 22 characters)\n", $name);
+    return;
+  }
+
+  # Warn about, but use, long rule names during --lint
+  if ($conf->{lint_rules}) {
+    if (length($name) > 50 && $name !~ /^__/ && $name !~ /^T_/) {
+      $self->lint_warn("config: warning: rule name '$name' is over 50 chars ".
+	     "(recommended maximum length is 22 characters)\n", $name);
+    }
+  }
+
   # all of these rule types are regexps
-  if ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_TESTS
-        || $type == $Mail::SpamAssassin::Conf::TYPE_BODY_TESTS
-        || $type == $Mail::SpamAssassin::Conf::TYPE_FULL_TESTS
-        || $type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_TESTS
-        || $type == $Mail::SpamAssassin::Conf::TYPE_URI_TESTS)
+  if ($type == $Mail::SpamAssassin::Conf::TYPE_BODY_TESTS ||
+      $type == $Mail::SpamAssassin::Conf::TYPE_FULL_TESTS ||
+      $type == $Mail::SpamAssassin::Conf::TYPE_RAWBODY_TESTS ||
+      $type == $Mail::SpamAssassin::Conf::TYPE_URI_TESTS)
   {
-    return unless $self->is_regexp_valid($name, $text);
+    return unless $self->is_delimited_regexp_valid($name, $text);
+  }
+  if ($type == $Mail::SpamAssassin::Conf::TYPE_HEAD_TESTS)
+  {
+    my ($pat) = ($text =~ /^\s*\S+\s*(?:\=|\!)\~\s*(\S.*?\S)\s*$/);
+    if ($pat) { $pat =~ s/\s+\[if-unset:\s+(.+)\]\s*$//; }
+    return unless $self->is_delimited_regexp_valid($name, $pat);
+  }
+  elsif ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS)
+  {
+    return unless $self->is_meta_valid($name, $text);
   }
 
   $conf->{tests}->{$name} = $text;
   $conf->{test_types}->{$name} = $type;
-  $conf->{tflags}->{$name} ||= '';
+  if ($type == $Mail::SpamAssassin::Conf::TYPE_META_TESTS) {
+    $conf->{priority}->{$name} ||= 500;
+  }
+  else {
+    $conf->{priority}->{$name} ||= 0;
+  }
   $conf->{priority}->{$name} ||= 0;
   $conf->{source_file}->{$name} = $self->{currentfile};
 
+  if ($self->{main}->{keep_config_parsing_metadata}) {
+    $conf->{if_stack}->{$name} = $self->get_if_stack_as_string();
+  }
+
+  # if we found this rule in a user_prefs file, it's a user rule -- note that
+  # we may need to recompile the rule code for this type (if they've already
+  # been compiled, e.g. in spamd).
+  #
+  # Note: the want_rebuild_for_type 'flag' is actually a counter; it is decremented
+  # after each scan.  This ensures that we always recompile at least once more;
+  # once to *define* the rule, and once afterwards to *undefine* the rule in the
+  # compiled ruleset again.
+  #
+  # If two consecutive scans use user rules, that's ok -- the second one will
+  # reset the counter, and we'll still recompile just once afterwards to undefine
+  # the rule again.
+  #
   if ($self->{scoresonly}) {
-    $conf->{user_rules_to_compile}->{$type} = 1;
+    $conf->{want_rebuild_for_type}->{$type} = 2;
     $conf->{user_defined_rules}->{$name} = 1;
   }
 }
@@ -761,18 +1086,100 @@ sub add_regression_test {
   }
 }
 
-sub is_regexp_valid {
-  my ($self, $name, $re) = @_;
-  if (eval { ("" =~ m{$re}); 1; }) {
-    return 1;
+sub is_meta_valid {
+  my ($self, $name, $rule) = @_;
 
-  } else {
-    warn "invalid regexp for rule $name: $re\n";
-    $self->{conf}->{errors}++;
+  my $meta = '';
+
+  # Lex the rule into tokens using a rather simple RE method ...
+  my $lexer = ARITH_EXPRESSION_LEXER;
+  my @tokens = ($rule =~ m/$lexer/g);
+  if (length($name) == 1) {
+    print "$name $_\n " for @tokens;
+  }
+  # Go through each token in the meta rule
+  foreach my $token (@tokens) {
+    # Numbers can't be rule names
+    if ($token =~ /^(?:\W+|[+-]?\d+(?:\.\d+)?)$/) {
+      $meta .= "$token ";
+    }
+    # Zero will probably cause more errors
+    else {
+      $meta .= "0 ";
+    }
+  }
+
+  my $evalstr = 'my $x = ' . $meta . '; 1;';
+  if (eval $evalstr) {
+    return 1;
+  }
+  if ($@) {
+    my $err = $@;
+    $err =~ s/\s+(?:at|near)\b.*//s;
+    $err =~ s/Illegal division by zero/division by zero possible/i;
+    $self->lint_warn("config: invalid expression for rule $name: \"$rule\": $err\n", $name);
     return 0;
   }
 }
 
+sub is_delimited_regexp_valid {
+  my ($self, $name, $re) = @_;
+
+  if (!$re || $re !~ /^\s*m?(\W).*(?:\1|>|}|\)|\])[a-z]*\s*$/) {
+    $re ||= '';
+    $self->lint_warn("config: invalid regexp for rule $name: $re: missing or invalid delimiters\n", $name);
+    return 0;
+  }
+  return $self->is_regexp_valid($name, $re);
+}
+
+sub is_regexp_valid {
+  my ($self, $name, $re) = @_;
+
+  # OK, try to remove any normal perl-style regexp delimiters at
+  # the start and end, and modifiers at the end if present,
+  # so we can validate those too.
+  my $origre = $re;
+  my $safere = $re;
+  my $mods = '';
+  if ($re =~ s/^m{//) {
+    $re =~ s/}([a-z]*)$//; $mods = $1;
+  }
+  elsif ($re =~ s/^m\(//) {
+    $re =~ s/\)([a-z]*)$//; $mods = $1;
+  }
+  elsif ($re =~ s/^m<//) {
+    $re =~ s/>([a-z]*)$//; $mods = $1;
+  }
+  elsif ($re =~ s/^m(\W)//) {
+    $re =~ s/\Q$1\E([a-z]*)$//; $mods = $1;
+  }
+  elsif ($re =~ s/^\/(.*)\/([a-z]*)$/$1/) {
+    $mods = $2;
+  }
+  else {
+    $safere = "m#".$re."#";
+  }
+
+  # now prepend the modifiers, in order to check if they're valid
+  if ($mods) {
+    $re = "(?".$mods.")".$re;
+  }
+
+  # note: this MUST use m/...${re}.../ in some form or another, ie.
+  # interpolation of the $re variable into a code regexp, in order to test the
+  # security of the regexp.  simply using ("" =~ $re) will NOT do that, and
+  # will therefore open a hole!
+  if (eval { ("" =~ m#${re}#); 1; }) {
+    return 1;
+  }
+
+  my $err = $@;
+  $err =~ s/ at .*? line \d.*$//;
+  $self->lint_warn("config: invalid regexp for rule $name: $origre: $err\n", $name);
+  return 0;
+}
+
 ###########################################################################
 
 sub add_to_addrlist {
@@ -785,7 +1192,7 @@ sub add_to_addrlist {
     $re =~ s/[\000\\\(]/_/gs;			# paranoia
     $re =~ s/([^\*\?_a-zA-Z0-9])/\\$1/g;	# escape any possible metachars
     $re =~ tr/?/./;				# "?" -> "."
-    $re =~ s/\*/\.\*/g;				# "*" -> "any string"
+    $re =~ s/\*+/\.\*/g;			# "*" -> "any string"
     $conf->{$singlelist}->{$addr} = "^${re}\$";
   }
 }
@@ -803,7 +1210,7 @@ sub add_to_addrlist_rcvd {
     $re =~ s/[\000\\\(]/_/gs;			# paranoia
     $re =~ s/([^\*\?_a-zA-Z0-9])/\\$1/g;	# escape any possible metachars
     $re =~ tr/?/./;				# "?" -> "."
-    $re =~ s/\*/\.\*/g;				# "*" -> "any string"
+    $re =~ s/\*+/\.\*/g;			# "*" -> "any string"
     $conf->{$listname}->{$addr}{re} = "^${re}\$";
     $conf->{$listname}->{$addr}{domain} = [ $domain ];
   }
@@ -832,18 +1239,51 @@ sub remove_from_addrlist_rcvd {
 sub fix_path_relative_to_current_file {
   my ($self, $path) = @_;
 
+  # the path may be specified as "~/foo", so deal with that
+  $path = $self->{conf}->{main}->sed_path($path);
+
   if (!File::Spec->file_name_is_absolute ($path)) {
     my ($vol, $dirs, $file) = File::Spec->splitpath ($self->{currentfile});
     $path = File::Spec->catpath ($vol, $dirs, $path);
-    dbg ("plugin: fixed relative path: $path");
+    dbg("config: fixed relative path: $path");
   }
   return $path;
 }
 
 ###########################################################################
 
-sub dbg { Mail::SpamAssassin::dbg (@_); }
-sub sa_die { Mail::SpamAssassin::sa_die (@_); }
+sub lint_warn {
+  my ($self, $msg, $rule, $iserror) = @_;
+
+  if (!defined $iserror) { $iserror = 1; }
+
+  if ($self->{conf}->{main}->{lint_callback}) {
+    $self->{conf}->{main}->{lint_callback}->(
+          msg => $msg,
+          rule => $rule,
+          iserror => $iserror
+        );
+  }
+  elsif ($self->{conf}->{lint_rules}) {
+    warn $msg."\n";
+  }
+  else {
+    info($msg);
+  }
+
+  if ($iserror) {
+    $self->{conf}->{errors}++;
+  }
+}
+
+###########################################################################
+
+sub get_if_stack_as_string {
+  my ($self) = @_;
+  return join ' ', map {
+    $_->{conditional}
+  } @{$self->{if_stack}};
+}
 
 ###########################################################################
 
