@@ -19,6 +19,7 @@ BEGIN {
   our $NO_SPAMC_EXE;
   our $SKIP_SPAMC_TESTS;
   our $SSL_AVAILABLE;
+  our $SKIP_SETUID_NOBODY_TESTS = 0;
 
 }
 
@@ -30,6 +31,12 @@ BEGIN {
 sub sa_t_init {
   my $tname = shift;
 
+  # optimisation -- don't setup spamd test parameters unless we're
+  # called "spamd_something" or "spamc_foo"
+  if ($tname !~ /spam[cd]/) {
+    $NO_SPAMD_REQUIRED = 1;
+  }
+
   if ($config{PERL_PATH}) {
     $perl_path = $config{PERL_PATH};
   }
@@ -46,18 +53,18 @@ sub sa_t_init {
   $perl_cmd .= " -w" if !defined($ENV{'TEST_PERL_WARN'})  or $ENV{'TEST_PERL_WARN'}  ne 'no';
 
   $scr = $ENV{'SCRIPT'};
-  $scr ||= "$perl_cmd ../spamassassin";
+  $scr ||= "$perl_cmd ../spamassassin.raw";
 
-  $spamd = "$perl_cmd ../spamd/spamd";
+  $spamd = "$perl_cmd ../spamd/spamd.raw";
 
   $spamc = $ENV{'SPAMC_SCRIPT'};
   $spamc ||= "../spamc/spamc";
 
   $salearn = $ENV{'SALEARN_SCRIPT'};
-  $salearn ||= "$perl_cmd ../sa-learn";
+  $salearn ||= "$perl_cmd ../sa-learn.raw";
 
   $spamdhost = $ENV{'SPAMD_HOST'};
-  $spamdhost ||= "localhost";
+  $spamdhost ||= "127.0.0.1";
   $spamdport = $ENV{'SPAMD_PORT'};
   $spamdport ||= probably_unused_spamd_port();
 
@@ -82,19 +89,34 @@ sub sa_t_init {
 
   read_config();
 
-  $NO_SPAMC_EXE = ($RUNNING_ON_WINDOWS &&
+  # if running as root, ensure "nobody" can write to it too
+  if ($> == 0) {
+    $tmp_dir_mode = 0777;
+    umask 022;  # ensure correct permissions on files and dirs created here
+    # Bug 5529 initial fix: For now don't run a test as root if it has a problem resuting from setuid nobody
+    # FIXME: Eventually we can actually test setuid nobody and accessing ./log to make this test more fine grained
+    #  and we can create an accessible temp dir that some of the tests can use. But for now just skip those tests.
+    $SKIP_SETUID_NOBODY_TESTS = 1;
+  } else {
+    $tmp_dir_mode = 0755;
+  }
+
+  if (!$NO_SPAMD_REQUIRED) {
+    $NO_SPAMC_EXE = ($RUNNING_ON_WINDOWS &&
                    !$ENV{'SPAMC_SCRIPT'} &&
                    !(-e "../spamc/spamc.exe"));
-  $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
+    $SKIP_SPAMC_TESTS = ($NO_SPAMC_EXE ||
                        ($RUNNING_ON_WINDOWS && !$ENV{'SPAMD_HOST'})); 
-  $SSL_AVAILABLE = ((!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
+    $SSL_AVAILABLE = ((!$SKIP_SPAMC_TESTS) &&  # no SSL test if no spamc
                     (!$SKIP_SPAMD_TESTS) &&  # or if no local spamd
                     (`$spamc -V` =~ /with SSL support/) &&
                     (`$spamd --version` =~ /with SSL support/));
+  }
   # do not remove prior test results!
   # rmtree ("log");
 
   mkdir ("log", 0755);
+  chmod (0755, "log"); # set in case log already exists with wrong permissions
 
   rmtree ("log/user_state");
   rmtree ("log/outputdir.tmp");
@@ -102,7 +124,7 @@ sub sa_t_init {
   rmtree ("log/test_rules_copy");
   mkdir ("log/test_rules_copy", 0755);
 
-  for $file (<../rules/*.cf>) {
+  for $file (<../rules/*.cf>, <../rules/*.pm>, <../rules/*.pre>) {
     $base = basename $file;
     copy ($file, "log/test_rules_copy/$base")
       or warn "cannot copy $file to log/test_rules_copy/$base";
@@ -130,7 +152,8 @@ sub sa_t_init {
   # create an empty .prefs file
   open (PREFS, ">>log/test_default.cf"); close PREFS;
 
-  mkdir("log/user_state",0755);
+  mkdir("log/user_state",$tmp_dir_mode);
+  chmod ($tmp_dir_mode, "log/user_state");  # unaffected by umask
 
   $home = $ENV{'HOME'};
   $home ||= $ENV{'WINDIR'} if (defined $ENV{'WINDIR'});
@@ -138,11 +161,15 @@ sub sa_t_init {
 
   $ENV{'TEST_DIR'} = $cwd;
   $testname = $tname;
+
+  $spamd_run_as_user = ($RUNNING_ON_WINDOWS || ($> == 0)) ? "nobody" : (getpwuid($>))[0] ;
 }
 
 # a port number between 32768 and 65535; used to allow multiple test
 # suite runs on the same machine simultaneously
 sub probably_unused_spamd_port {
+  return 0 if $NO_SPAMD_REQUIRED;
+
   my $port;
   my @nstat = ();
   if (open(NSTAT, "netstat -a -n 2>&1 |")) {
@@ -229,11 +256,8 @@ sub sarun {
   my $post_redir = '';
   $args =~ s/ 2\>\&1$// and $post_redir = ' 2>&1';
 
-  rmtree ("log/outputdir.tmp"); # some tests use this
-  mkdir ("log/outputdir.tmp", 0755);
-
-  %found = ();
-  %found_anti = ();
+  recreate_outputdir_tmp();
+  clear_pattern_counters();
 
   if (defined $ENV{'SA_ARGS'}) {
     $args = $ENV{'SA_ARGS'} . " ". $args;
@@ -244,10 +268,11 @@ sub sarun {
   my $scrargs = "$scr $args";
   $scrargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$scrargs\n");
-  system ("$scrargs > log/$testname.${Test::ntest} $post_redir");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$scrargs > log/d.$testname/${Test::ntest} $post_redir");
   $sa_exitcode = ($?>>8);
   if ($sa_exitcode != 0) { return undef; }
-  &checkfile ("$testname.${Test::ntest}", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/${Test::ntest}", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -263,8 +288,7 @@ sub salearnrun {
   my $args = shift;
   my $read_sub = shift;
 
-  rmtree ("log/outputdir.tmp"); # some tests use this
-  mkdir ("log/outputdir.tmp", 0755);
+  recreate_outputdir_tmp();
 
   %found = ();
   %found_anti = ();
@@ -278,10 +302,11 @@ sub salearnrun {
   my $salearnargs = "$salearn $args";
   $salearnargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
   print ("\t$salearnargs\n");
-  system ("$salearnargs > log/$testname.${Test::ntest}");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$salearnargs > log/d.$testname/${Test::ntest}");
   $salearn_exitcode = ($?>>8);
   if ($salearn_exitcode != 0) { return undef; }
-  &checkfile ("$testname.${Test::ntest}", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/${Test::ntest}", $read_sub) if (defined $read_sub);
   1;
 }
 
@@ -291,11 +316,15 @@ sub scrun {
 sub scrunwithstderr {
   spamcrun (@_, 1);
 }
+sub scrunwantfail {
+  spamcrun (@_, 1, 1);
+}
 
 sub spamcrun {
   my $args = shift;
   my $read_sub = shift;
   my $capture_stderr = shift;
+  my $expect_failure = shift;
 
   if (defined $ENV{'SC_ARGS'}) {
     $args = $ENV{'SC_ARGS'} . " ". $args;
@@ -317,20 +346,27 @@ sub spamcrun {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs\n");
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
   if ($capture_stderr) {
-    system ("$spamcargs > log/$testname.out 2>&1");
+    system ("$spamcargs > log/d.$testname/out.${Test::ntest} 2>&1");
   } else {
-    system ("$spamcargs > log/$testname.out");
+    system ("$spamcargs > log/d.$testname/out.${Test::ntest}");
   }
 
   $sa_exitcode = ($?>>8);
-  if ($sa_exitcode != 0) { stop_spamd(); return undef; }
+  if (!$expect_failure) {
+    if ($sa_exitcode != 0) { stop_spamd(); return undef; }
+  }
 
   %found = ();
   %found_anti = ();
-  &checkfile ("$testname.out", $read_sub) if (defined $read_sub);
+  &checkfile ("d.$testname/out.${Test::ntest}", $read_sub) if (defined $read_sub);
 
-  ($sa_exitcode == 0);
+  if ($expect_failure) {
+    ($sa_exitcode != 0);
+  } else {
+    ($sa_exitcode == 0);
+  }
 }
 
 sub spamcrun_background {
@@ -353,7 +389,8 @@ sub spamcrun_background {
   $spamcargs =~ s!/!\\!g if ($^O =~ /^MS(DOS|Win)/i);
 
   print ("\t$spamcargs &\n");
-  system ("$spamcargs > log/$testname.bg &") and return 0;
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  system ("$spamcargs > log/d.$testname/bg.${Test::ntest} &") and return 0;
 
   1;
 }
@@ -370,16 +407,22 @@ sub sdrun {
   1;
 }
 
-sub start_spamd {
+sub recreate_outputdir_tmp {
+  rmtree ("log/outputdir.tmp"); # some tests use this
+  mkdir ("log/outputdir.tmp", $tmp_dir_mode);
+  chmod ($tmp_dir_mode, "log/outputdir.tmp");  # unaffected by umask
+}
 
+# out: $spamd_stderr
+sub start_spamd {
+  die "NO_SPAMD_REQUIRED in start_spamd! oops" if $NO_SPAMD_REQUIRED;
   return if $SKIP_SPAMD_TESTS;
 
   my $spamd_extra_args = shift;
 
   return if (defined($spamd_pid) && $spamd_pid > 0);
 
-  rmtree ("log/outputdir.tmp"); # some tests use this
-  mkdir ("log/outputdir.tmp", 0755);
+  recreate_outputdir_tmp();
 
   if (defined $ENV{'SD_ARGS'}) {
     $spamd_extra_args = $ENV{'SD_ARGS'} . " ". $spamd_extra_args;
@@ -408,14 +451,21 @@ sub start_spamd {
       qq{-p}, $spamdport,
     );
   }
+  if ($spamd_extra_args !~ /(?:--socketpath)/) {
+    push(@spamd_args,
+      qq{-A}, $spamdhost,
+    );
+  }
 
   if ($set_test_prefs) {
     warn "oops! SATest.pm: a test prefs file was created, but spamd isn't reading it\n";
   }
 
-  my $spamd_stdout = "log/$testname-spamd.out";
-  my $spamd_stderr = "log/$testname-spamd.err";
-  my $spamd_stdlog = "log/$testname-spamd.log";
+  (-d "log/d.$testname") or mkdir ("log/d.$testname", 0755);
+  my $spamd_stdout = "log/d.$testname/spamd.out.${Test::ntest}";
+     $spamd_stderr = "log/d.$testname/spamd.err.${Test::ntest}";    #  global
+  my $spamd_stdlog = "log/d.$testname/spamd.log.${Test::ntest}";
+
   my $spamd_forker = $ENV{'SPAMD_FORKER'}   ?
                        $ENV{'SPAMD_FORKER'} :
                      $RUNNING_ON_WINDOWS    ?
@@ -425,11 +475,12 @@ sub start_spamd {
                        $spamd_forker,
                        qq{SATest.pl},
                        qq{-Mredirect},
-                       qq{-o${spamd_stdout}},
                        qq{-O${spamd_stderr}},
+                       qq{-o${spamd_stdout}},
                        qq{--},
                        @spamd_args,
                        $spamd_extra_args,
+                       qq{-s ${spamd_stderr}.timestamped},
                        qq{&},
                     );
   unlink ($spamd_stdout, $spamd_stderr, $spamd_stdlog);
@@ -450,6 +501,12 @@ sub start_spamd {
         # Yes, DO retry on this error. I'm getting test failures otherwise
         # /Address already in use/ and $retries = 0;
 	/server pid: (\d+)/ and $spamd_pid = $1;
+
+        if (/ERROR/) {
+          warn "spamd error! $_";
+          $retries = 0; last;
+        }
+
 	$spamdlog .= $_;
       }
       close IN;
@@ -468,6 +525,7 @@ sub start_spamd {
 }
 
 sub stop_spamd {
+  die "NO_SPAMD_REQUIRED in stop_spamd! oops" if $NO_SPAMD_REQUIRED;
   return 0 if ( defined($spamd_already_killed) || $SKIP_SPAMD_TESTS);
 
   $spamd_pid ||= 0;
@@ -527,7 +585,14 @@ sub checkfile {
 
   # print "Checking $filename\n";
   if (!open (IN, "< log/$filename")) {
-    warn "cannot open log/$filename"; return undef;
+    # could be it already contains the "log/" prefix?
+    if (!open (IN, "< $filename")) {
+      warn "cannot open log/$filename or $filename"; return undef;
+    } else {
+      push @files_checked, "$filename";
+    }
+  } else {
+    push @files_checked, "log/$filename";
   }
   &$read_sub();
   close IN;
@@ -537,6 +602,11 @@ sub checkfile {
 
 sub pattern_to_re {
   my $pat = shift;
+
+  if ($pat =~ /^\/(.*)\/$/) {
+    return $1;
+  }
+
   $pat = quotemeta($pat);
 
   # make whitespace irrelevant; match any amount as long as the
@@ -582,14 +652,22 @@ sub patterns_run_cb {
 }
 
 sub ok_all_patterns {
+  my ($dont_ok) = shift;
+
+  my $wasfailure = 0;
   foreach my $pat (sort keys %patterns) {
     my $type = $patterns{$pat};
     print "\tChecking $type\n";
     if (defined $found{$type}) {
-      ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      if (!$dont_ok) {
+        ok ($found{$type} == 1) or warn "Found more than once: $type\n";
+      }
     } else {
       warn "\tNot found: $type = $pat\n";
-      ok (0);                     # keep the right # of tests
+      if (!$dont_ok) {
+        ok (0);                     # keep the right # of tests
+      }
+      $wasfailure++;
     }
   }
   foreach my $pat (sort keys %anti_patterns) {
@@ -597,13 +675,21 @@ sub ok_all_patterns {
     print "\tChecking for anti-pattern $type\n";
     if (defined $found_anti{$type}) {
       warn "\tFound anti-pattern: $type = $pat\n";
-      ok (0);
+      if (!$dont_ok) { ok (0); }
+      $wasfailure++;
     }
     else
     {
-      ok (1);
+      if (!$dont_ok) { ok (1); }
     }
   }
+
+  if ($wasfailure) {
+    warn "Output can be examined in: ".join(' ', @files_checked)."\n";
+    return 0;
+  } else {
+    return 1;
+  }
 }
 
 sub skip_all_patterns {
@@ -640,6 +726,7 @@ sub skip_all_patterns {
 sub clear_pattern_counters {
   %found = ();
   %found_anti = ();
+  @files_checked = ();
 }
 
 sub read_config {
@@ -715,4 +802,72 @@ sub cleanup_safe_tmpdir {
   }
 }
 
+sub wait_for_file_to_change_or_disappear {
+  my ($f, $timeout, $action) = @_;
+
+  my $lastmod = (-M $f);
+
+  $action->();
+
+  my $wait = 0;
+  my $newlastmod;
+  do {
+    sleep (int($wait++ / 4) + 1) if $timeout > 0;
+    $timeout--;
+    $newlastmod = (-M $f);
+  } while((-e $f) && defined($newlastmod) &&
+                $newlastmod == $lastmod && $timeout);
+}
+
+sub wait_for_file_to_appear {
+  my ($f, $timeout) = @_;
+
+  # note that the wait period increases the longer it takes,
+  # 20 retries works out to a total of 60 seconds
+  my $wait = 0;
+  do {
+    sleep (int($wait++ / 4) + 1) if $timeout > 0;
+    $timeout--;
+  } while((!-e $f || -z $f) && $timeout);
+}
+
+sub read_from_pidfile {
+  my $f = shift;
+  my $npid = 0;
+  my $retries = 5;
+
+  do {
+    if ($retries != 5) {
+      sleep 1;
+      warn "retrying read of pidfile $f, due to short/nonexistent read: ".
+            "retry $retries";
+    }
+    $retries--;
+
+    if (!open (PID, "<".$f)) {
+      warn "Could not open pid file ${f}: $!\n";     # and retry
+      next;
+    }
+
+    $npid = <PID>;
+    if (defined $npid) { chomp $npid; }
+    close(PID);
+
+    if (!$npid || $npid < 1) {
+      warn "failed to read anything sensible from $f, retrying read";
+      $npid = 0;
+      next;
+    }
+    if (!kill (0, $npid)) {
+      warn "failed to kill -0 $npid, retrying read";
+      $npid = 0;
+    }
+
+  } until ($npid > 1 or $retries == 0);
+
+  return $npid;
+}
+
+sub dbgprint { print STDOUT "[".time()."] ".$_[0]; }
+
 1;
